<?php
/**
 * A general file-based stack class
 *
 * @author		Roland Haeder <webmaster@ship-simu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2013 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.ship-simu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
class BaseFileStack extends BaseStacker {
	/**
	 * Magic for this stack
	 */
	const STACK_MAGIC = 'STACKv0.1';

	/**
	 * Separator for header data
	 */
	const SEPARATOR_HEADER_DATA = 0x01;

	/**
	 * Separator header->entries
	 */
	const SEPARATOR_HEADER_ENTRIES = 0x02;

	/**
	 * Separator hash->name
	 */
	const SEPARATOR_HASH_NAME = 0x03;

	/**
	 * Length of name
	 */
	const LENGTH_NAME = 10;

	/**
	 * Length of count
	 */
	const LENGTH_COUNT = 20;

	/**
	 * Length of position
	 */
	const LENGTH_POSITION = 20;

	/**
	 * Counter for total entries
	 */
	private $totalEntries = 0;

	/**
	 * Current seek position
	 */
	private $seekPosition = 0;

	/**
	 * Size of header
	 */
	private $headerSize = 0;

	/**
	 * File header
	 */
	private $header = array();

	/**
	 * Protected constructor
	 *
	 * @param	$className	Name of the class
	 * @return	void
	 */
	protected function __construct ($className) {
		// Call parent constructor
		parent::__construct($className);

		// Calculate header size
		$this->headerSize =
			strlen(self::MAGIC) +
			strlen(self::SEPARATOR_HEADER_DATA) +
			self::LENGTH_COUNT +
			strlen(self::SEPARATOR_HEADER_DATA) +
			self::LENGTH_POSITION +
			strlen(self::SEPARATOR_HEADER_ENTRIES)
		);
	}

	/**
	 * Getter for total entries
	 *
	 * @return	$totalEntries	Total entries in this stack
	 */
	private function getCounter () {
		// Get it
		return $this->totalEntries;
	}

	/**
	 * Increment counter
	 *
	 * @return	void
	 */
	private function incrementCounter () {
		// Get it
		$this->totalEntries++;
	}

	/**
	 * Getter for seek position
	 *
	 * @return	$seekPosition	Current seek position (stored here in object)
	 */
	private function getSeekPosition () {
		// Get it
		return $this->seekPosition;
	}

	/**
	 * Setter for seek position
	 *
	 * @param	$seekPosition	Current seek position (stored here in object)
	 * @return	void
	 */
	private function setSeekPosition ($seekPosition) {
		// And set it
		$this->seekPosition = $seekPosition;
	}

	/**
	 * Updates seekPosition attribute from file to avoid to much access on file.
	 *
	 * @return	void
	 */
	private function updateSeekPosition () {
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] CALLED!', __METHOD__, __LINE__));

		// Get key (= seek position)
		$seekPosition = $this->getIteratorInstance()->key();
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Setting seekPosition=%s', __METHOD__, __LINE__, $seekPosition));

		// And set it here
		$this->setSeekPosition($seekPosition);

		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] EXIT!', __METHOD__, __LINE__));
	}

	/**
	 * Reads the file header
	 *
	 * @return	void
	 */
	private function readFileHeader () {
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] CALLED!', __METHOD__, __LINE__));

		// First rewind to beginning as the header sits at the beginning ...
		$this->getIteratorInstance()->rewind();

		// Then read it (see constructor for calculation)
		$data = $this->getIteratorInstance()->read($this->headerLength);

		// Last character must be the separator
		assert(substr($data, -1, 1) == self::SEPARATOR_HEADER_ENTRIES);

		// Okay, then remove it
		$data = substr($data, 0, -1);

		// And update seek position
		$this->updateSeekPosition();

		/*
		 * Now split it:
		 *
		 * 0 => Magic
		 * 1 => Total entries
		 * 2 => Current seek position
		 */
		$this->header = explode(self::SEPARATOR_HEADER_DATA, $data);

		// Check if the array has only 3 elements
		assert(count($this->header) == 3);

		// Check magic
		assert($this->header[0] == self::MAGIC);

		// Check length of count and seek position
		assert(strlen($this->header[1]) == self::LENGTH_COUNT);
		assert(strlen($this->header[2]) == self::LENGTH_POSITION);

		// Decode count and seek position
		$this->header[1] = hex2bin($this->header[1]);
		$this->header[2] = hex2bin($this->header[2]);

		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] EXIT!', __METHOD__, __LINE__));
	}

	/**
	 * Checks whether the file header is initialized
	 *
	 * @return	$isInitialized	Whether the file header is initialized
	 */
	private function isFileHeaderInitialized () {
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] CALLED!', __METHOD__, __LINE__));
		// Default is not initialized
		$isInitialized = FALSE;

		// Is the file initialized?
		if ($this->isFileInitialized()) {
			// Some bytes has been written, so rewind to start of it.
			$rewindStatus = $this->getIteratorInstance()->rewind();
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] rewindStatus=%s', __METHOD__, __LINE__, $rewindStatus));

			// Is the rewind() call successfull?
			if ($rewindStatus != 1) {
				// Something bad happened
				self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Could not rewind().', __METHOD__, __LINE__));
			} // END - if

			// Read file header
			$this->readFileHeader();

			// Check if its format is valid.
		} // END - if

		// Return result
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] isInitialized=%d - EXIT!', __METHOD__, __LINE__, intval($isInitialized)));
		return $isInitialized;
	}

	/**
	 * Checks whether the file-based stack has been initialized
	 *
	 * @return	$isInitialized		Whether the file's size is zero
	 */
	private function isFileInitialized () {
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] CALLED!', __METHOD__, __LINE__));

		// Get it from iterator which holds the pointer instance. If FALSE is returned
		$fileSize = $this->getIteratorInstance()->size();
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] fileSize=%s', __METHOD__, __LINE__, $fileSize));

		/*
		 * The returned file size should not be FALSE or NULL as this means
		 * that the pointer class does not work correctly.
		 */
		assert(is_int($fileSize));

		// Is more than 0 returned?
		$isInitialized = ($fileSize > 0);

		// Return result
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] isInitialized=%d - EXIT!', __METHOD__, __LINE__, intval($isInitialized)));
		return $isInitialized;
	}

	/**
	 * Creates the file-stack's header
	 *
	 * @return	void
	 */
	private function createFileHeader () {
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] CALLED!', __METHOD__, __LINE__));
		// The file's header should not be initialized here
		assert(!$this->isFileHeaderInitialized());

		// Simple flush file header which will create it.
		$this->flushFileHeader();

		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] EXIT!!', __METHOD__, __LINE__));
	}

	/**
	 * Flushes the file header
	 *
	 * @return	void
	 */
	private function flushFileHeader () {
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] CALLED!', __METHOD__, __LINE__));

		// Put all informations together
		$header = sprintf('%s%s%s%s%s%s',
			// Magic
			self::STACK_MAGIC,

			// Separator magic<->count
			chr(self::SEPARATOR_HEADER_DATA),

			// Total entries (will be zero) and pad it to 20 chars
			str_pad($this->dec2hex($this->getCounter()), self::LENGTH_COUNT, '0', STR_PAD_LEFT),

			// Separator count<->seek position
			chr(self::SEPARATOR_HEADER_DATA),

			// Position (will be zero)
			str_pad($this->dec2hex($this->getSeekPosition(), 2), self::LENGTH_POSITION, '0', STR_PAD_LEFT),

			// Separator position<->entries
			chr(self::SEPARATOR_HEADER_ENTRIES)
		);

		// Write it to disk (header is always at seek position 0)
		$this->writeData(0, $header);

		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] EXIT!', __METHOD__, __LINE__));
	}

	/**
	 * Writes data at given position
	 *
	 * @param	$seekPosition	Seek position
	 * @param	$data			Data to be written
	 * @return	void
	 */
	private function writeData ($seekPosition, $data) {
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] seekPosition=%s,data()=%s - CALLED!', __METHOD__, __LINE__, $seekPosition, strlen($data)));

		// Write data at given position
		$this->getIteratorInstance()->writeAtPosition($seekPosition, $data);

		// Update seek position
		$this->updateSeekPosition();

		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] EXIT!', __METHOD__, __LINE__));
	}

	/**
	 * Pre-allocates file (if enabled) with some space for later faster write access.
	 *
	 * @return	void
	 */
	private function preAllocateFile () {
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] CALLED!', __METHOD__, __LINE__));

		// Is it enabled?
		if ($this->getConfigInstance()->getConfigEntry('file_stack_pre_allocate_enabled') != 'Y') {
			// Not enabled
			self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Not pre-allocating stack file.', __METHOD__, __LINE__));

			// Don't continue here.
			return;
		} // END - if

		// Message to user
		self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Pre-allocating stack file ...', __METHOD__, __LINE__));

		/*
		 * Calculate minimum length for one entry:
		 * minimum length = hash length + separator + name + minimum entry size = ?? + 1 + 10 + 1 = ??
		 */
		$minLengthEntry = self::getHashLength() + strlen(self::SEPARATOR_HASH_NAME) + self::LENGTH_NAME + 1;
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] minLengthEntry=%s', __METHOD__, __LINE__, $minLengthEntry));

		// Calulcate seek position
		$seekPosition = $minLengthEntry * $this->getConfigInstance()->getConfigEntry('file_stack_pre_allocate_count');
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] seekPosition=%s', __METHOD__, __LINE__, $seekPosition));

		// Now simply write a NUL there. This will pre-allocate the file.
		$this->writeData($seekPosition, chr(0));

		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] EXIT!', __METHOD__, __LINE__));
	}

	/**
	 * Initializes this file-based stack.
	 *
	 * @param	$fileName	File name of this stack
	 * @return	void
	 */
	protected function initFileStack ($fileName) {
		// Get a file i/o pointer instance
		$pointerInstance = ObjectFactory::createObjectByConfiguredName('file_raw_input_output_class', array($fileName));

		// Get iterator instance
		$iteratorInstance = ObjectFactory::createObjectByConfiguredName('file_io_iterator_class', array($pointerInstance));

		// Is the instance implementing the right interface?
		assert($iteratorInstance instanceof SeekableWritableFileIterator);

		// Set iterator here
		$this->setIteratorInstance($iteratorInstance);

		// Is the file's header initialized?
		if ($this->isFileHeaderInitialized()) {
			// Then load it
			$this->loadFileHeader();
		} else {
			// No, then create it (which may pre-allocate the stack)
			$this->createFileHeader();

			// And pre-allocate a bit
			$this->preAllocateFile();
		}
	}

	/**
	 * Initializes given stacker
	 *
	 * @param	$stackerName	Name of the stack
	 * @param	$forceReInit	Force re-initialization
	 * @return	void
	 * @throws	AlreadyInitializedStackerException	If the stack is already initialized
	 */
	public function initStack ($stackerName, $forceReInit = FALSE) {
		// Is the stack already initialized?
		if (($forceReInit === FALSE) && ($this->isStackInitialized($stackerName))) {
			// Then throw the exception
			throw new AlreadyInitializedStackerException(array($this, $stackerName, $forceReInit), self::EXCEPTION_STACKER_ALREADY_INITIALIZED);
		} // END - if

		// Initialize the given stack
		$this->partialStub('stackerName=' . $stackerName . ',forceReInit=' . intval($forceReInit));
	}

	/**
	 * Checks whether the given stack is initialized (set in array $stackers)
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$isInitialized	Whether the stack is initialized
	 */
	public function isStackInitialized ($stackerName) {
		// Is is there?
		$this->partialStub('stackerName=' . $stackerName);
		$isInitialized = TRUE;

		// Return result
		return $isInitialized;
	}

	/**
	 * Getter for size of given stack (array count)
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$count			Size of stack (array count)
	 * @throws	NoStackerException	If given stack is missing
	 */
	public function getStackCount ($stackerName) {
		// Is the stack not yet initialized?
		if (!$this->isStackInitialized($stackerName)) {
			// Throw an exception
			throw new NoStackerException(array($this, $stackerName), self::EXCEPTION_NO_STACKER_FOUND);
		} // END - if

		// Now, count the array of entries
		$this->partialStub('stackerName=' . $stackerName);
		$count = 0;

		// Return result
		return $count;
	}

	/**
	 * Adds a value to given stack
	 *
	 * @param	$stackerName	Name of the stack
	 * @param	$value			Value to add to this stacker
	 * @return	void
	 * @throws	FullStackerException	Thrown if the stack is full
	 */
	protected function addValue ($stackerName, $value) {
		// Is the stack not yet initialized or full?
		if (!$this->isStackInitialized($stackerName)) {
			// Then do it here
			$this->initStack($stackerName);
		} elseif ($this->isStackFull($stackerName)) {
			// Stacker is full
			throw new FullStackerException(array($this, $stackerName, $value), self::EXCEPTION_STACKER_IS_FULL);
		}

		// Now add the value to the stack
		$this->partialStub('stackerName=' . $stackerName . ',value[]=' . gettype($value));
	}

	/**
	 * Get last value from named stacker
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$value			Value of last added value
	 * @throws	NoStackerException	If the named stacker was not found
	 * @throws	EmptyStackerException	If the named stacker is empty
	 */
	protected function getLastValue ($stackerName) {
		// Is the stack not yet initialized or full?
		if (!$this->isStackInitialized($stackerName)) {
			// Throw an exception
			throw new NoStackerException(array($this, $stackerName), self::EXCEPTION_NO_STACKER_FOUND);
		} elseif ($this->isStackEmpty($stackerName)) {
			// Throw an exception
			throw new EmptyStackerException(array($this, $stackerName), self::EXCEPTION_STACKER_IS_EMPTY);
		}

		// Now get the last value
		$this->partialStub('stackerName=' . $stackerName);
		$value = NULL;

		// Return it
		return $value;
	}

	/**
	 * Get first value from named stacker
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$value			Value of last added value
	 * @throws	NoStackerException	If the named stacker was not found
	 * @throws	EmptyStackerException	If the named stacker is empty
	 */
	protected function getFirstValue ($stackerName) {
		// Is the stack not yet initialized or full?
		if (!$this->isStackInitialized($stackerName)) {
			// Throw an exception
			throw new NoStackerException(array($this, $stackerName), self::EXCEPTION_NO_STACKER_FOUND);
		} elseif ($this->isStackEmpty($stackerName)) {
			// Throw an exception
			throw new EmptyStackerException(array($this, $stackerName), self::EXCEPTION_STACKER_IS_EMPTY);
		}

		// Now get the first value
		$this->partialStub('stackerName=' . $stackerName);
		$value = NULL;

		// Return it
		return $value;
	}

	/**
	 * "Pops" last entry from stack
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$value			Value "poped" from array
	 * @throws	NoStackerException	If the named stacker was not found
	 * @throws	EmptyStackerException	If the named stacker is empty
	 */
	protected function popLast ($stackerName) {
		// Is the stack not yet initialized or full?
		if (!$this->isStackInitialized($stackerName)) {
			// Throw an exception
			throw new NoStackerException(array($this, $stackerName), self::EXCEPTION_NO_STACKER_FOUND);
		} elseif ($this->isStackEmpty($stackerName)) {
			// Throw an exception
			throw new EmptyStackerException(array($this, $stackerName), self::EXCEPTION_STACKER_IS_EMPTY);
		}

		// Now, remove the last entry, we don't care about the return value here, see elseif() block above
		$this->partialStub('stackerName=' . $stackerName);
		return NULL;
	}

	/**
	 * "Pops" first entry from stack
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$value			Value "shifted" from array
	 * @throws	NoStackerException	If the named stacker was not found
	 * @throws	EmptyStackerException	If the named stacker is empty
	 */
	protected function popFirst ($stackerName) {
		// Is the stack not yet initialized or full?
		if (!$this->isStackInitialized($stackerName)) {
			// Throw an exception
			throw new NoStackerException(array($this, $stackerName), self::EXCEPTION_NO_STACKER_FOUND);
		} elseif ($this->isStackEmpty($stackerName)) {
			// Throw an exception
			throw new EmptyStackerException(array($this, $stackerName), self::EXCEPTION_STACKER_IS_EMPTY);
		}

		// Now, remove the last entry, we don't care about the return value here, see elseif() block above
		$this->partialStub('stackerName=' . $stackerName);
		return NULL;
	}
}

// [EOF]
?>
